/*==========================================================
 * Program : gsmts0338.c                   Project : smslink
 * Authors : Philippe Andersson.
 *           Franois Baligant <francois@euronet.be>.
 * Date    : 18/10/01
 * Version : 0.01a
 * Notice  : (c) Les Ateliers du Heron, 1998 for Scitex Europe, S.A.
 * Comment : PDU Manipulation tool. Encodes and decodes GSM TS
 *           03.38 alphabet.
 *
 * Modification History :
 * - 0.01a (18/10/01) : Initial release.
 *========================================================*/
#include <unistd.h>
#include <stdio.h>                         /* for fprintf */
#include <stdlib.h>                  /* for errno & stuff */
#include <errno.h>
#include <string.h>

#include "sms_serv.h"
#include "pdu.h"
#include "gsmts0338.h"

/*========================================================*/
/**********             LOCAL DEFINES              ********/
/*========================================================*/

/*========================================================*/
/**********           GLOBAL VARIABLES             ********/
/*========================================================*/
int debug = FALSE;

/*========================================================*/
/**********           MAIN PROGRAM LOOP            ********/
/*========================================================*/
int main (int argc, char *argv[])
{
  extern char *optarg;
  extern int optind, optopt;
  int c;                /* option char returned by getopt */
  int errflag = 0;       /* error in command line options */
  int task = 0;
  char *string;
  char *message;

  /*--------------------------------------Initialisations */
  string = (char *) malloc ((BIGBUFF + 1) * sizeof (char));
  if (!string) {
    fprintf (stderr, "gsmts0338: main(): can't malloc string\n");
    exit (1);
  }
  string[0] = '\0';
  message = (char *) malloc ((BIGBUFF + 1) * sizeof (char));
  if (!message) {
    fprintf (stderr, "gsmts0338: main(): can't malloc message\n");
    exit (1);
  }
  message[0] = '\0';

  /*-------------------------Command line processing loop */
  while ((c = getopt (argc, argv, ":ghvd:e:")) != -1) {
    switch (c) {
      case 'g':                               /* debug ON */
        debug = TRUE;
        break;
	
      case 'd':                                 /* decode */
        if (strlen (optarg) > BIGBUFF) {
	  fprintf (stderr, "gsmts0338: -d parameter too long, max. is %d\n",
	          BIGBUFF);
	  errflag++;
	}
	else
	  strcpy (string, optarg);
	task = GSMTS_DECODE;
        break;

      case 'e':                                 /* encode */
        if (strlen (optarg) > BIGBUFF) {
	  fprintf (stderr, "gsmts0338: -e parameter too long, max. is %d\n",
	          BIGBUFF);
	  errflag++;
	}
	else
          strcpy (message, optarg);
	task = GSMTS_ENCODE;
        break;

      case 'v':                           /* version info */
	printf ("gsmts0338 ver. %s (%s), SMSLink PDU Manipulation Tool.\n", 
        	GSMTS_VERSION, GSMTS_DATE);
	printf ("(c) Les Ateliers du Heron, 1998-2001.\n");
        exit (0);
        break;
	
      case 'h':                             /* usage help */
	printf ("\ngsmts0338 ver. %s (%s), SMSLink PDU Manipulation Tool.\n", 
        	GSMTS_VERSION, GSMTS_DATE);
	printf ("(c) Les Ateliers du Heron, 1998-2001.\n");
	printf ("\nUsage:\n");
	printf ("gsmts0338 [-gvh] (-d string | -e msg)\n");
	printf ("\nWhere: string  = GSM TS 03.38 string to decode\n");
	printf ("       msg     = message to encode to GSM TS 03.38\n");
	printf ("Other options:\n");
	printf ("-g = sets debugging ON [def. = OFF]\n");
	printf ("-v = prints version information and exits\n");
	printf ("-h = prints this usage information and exits\n");
        exit (0);
        break;
	
      case ':':             /* parameter without argument */
        fprintf (stderr, "gsmts0338: Option -%c requires an argument.\n",
                optopt);
        errflag++;
        break;

      case '?':               /* unrecognized option char */
        fprintf (stderr, "gsmts0338: Unrecognized option: -%c.\n", optopt);
        errflag++;
        break;

      default:
        fprintf (stderr, "gsmts0338: getopt() returned char. code 0%o ??\n", c);
    }                                       /* switch (c) */
  }                                       /* while (c...) */
  /*----------------------Check for required parameter(s) */
  /* either -d or -e */
  if (!errflag) {
    if (task == 0) {
      fprintf (stderr, "gsmts0338: You must provide at least one of -d or -e.\n");
      errflag++;
    }
  }                                      /* if (!errflag) */

  /*-------------------Check for conflicting parameter(s) */
  /* -d with -e */
  if ((strlen (message) > 0) && (strlen (string) > 0)) {
      fprintf (stderr, "gsmts0338: Conflicting parameters -d and -e.\n");
      errflag++;
  }
  /*-----------------------On error, display usage & exit */
  if (errflag) {
    fprintf (stderr, "\n==> gsmts0338 -h for usage details.\n");
    exit (2);
  }
  /* DEBUG ...............................................*/
  if (debug) {
    fprintf (stderr, "Debugging mode is turned ON.\n");
    fprintf (stderr, "Those are the current parameters :\n");
    fprintf (stderr, "string   = [%s]\n", string);
    fprintf (stderr, "message  = [%s]\n", message);
  }

  /*---------------------------Start real processing loop */
  switch (task) {
    case GSMTS_ENCODE :
      printf ("Original message: [%s] (%d char.)\n", message, strlen (message));
      ascii7topacked8 (message, string);
      printf ("Encoded string: [%s] (%d char.)\n", string, strlen (string));
      break;
      
    case GSMTS_DECODE :
      printf ("Encoded string: [%s] (%d char.)\n", string, strlen (string));
      packed8toascii7 (string, message);
      printf ("Decoded message: [%s] (%d char.)\n", message, strlen (message));
      break;
      
    default :
      fprintf (stderr, "gsmts0338: Unsupported task #%d, internal error.\n", task);
      exit (1);
      break;
      
  }                                      /* switch (task) */


  /*------------------------------------------Conclusions */
  /* free what's need to be */
  free (string);
  free (message);

  /*------------------------------------------End program */
  exit (0);
}

/*==========================================================
 * EOF : gsmts0338.c
 *===================*/
