/*==========================================================
 * Program : pdu.c                         Project : smslink
 * Authors : Philippe Andersson.
 *           Franois Baligant <francois@euronet.be>.
 * Date    : 18/10/01
 * Version : 0.13b
 * Notice  : (c) Les Ateliers du Heron, 1998 for Scitex Europe, S.A.
 * Comment : Library of PDU-related functions for the smslink servers.
 *
 * Modification History :
 * - 0.01a (07/04/00) : Initial release.
 * - 0.02a (13/04/00) : Added the ascii7topacked8() function.
 * - 0.03a (26/04/00) : Fixed a misunderstanding in encode_pdu().
 *   TP_UDL should be measured in chars, not in bytes. Thanks
 *   to <eric.guidi@wavecom.fr> for pointing that out.
 * - 0.04a (08/05/00) : Fixed a second misunderstanding (pad
 *   the TP_DA before byteswap, not after. Thanks again
 *   to <eric.guidi@wavecom.fr> for pointing that out.
 * ++++ Switch to Beta ++++
 * - 0.05b (20/11/00) : Included the packed8toascii7() function.
 * - 0.06b (21/11/00) : Added the decode_pdu() function.
 * - 0.07b (22/11/00) : Added the parse_csts_date() and
 *   parse_csts_time() functions.
 * - 0.08b (23/11/00) : Included cap_matrix in the parameters
 *   passed to decode_pdu(), in order to handle those devices
 *   that have the SCA at the start of the PDU string.
 * - 0.09b (24/11/00) : Fixed a bug in the decode_pdu() function
 *   for cases where HAVE_SCA_IN_PDU is true. <scalen> is not
 *   coded the same as <oalen>. Simplified <tpdu> structure in
 *   encode_pdu() function (caused problems for PCFF900). 
 * - 0.10b (18/02/01) : Modified packed8toascii7() so that it
 *   would return in error in case a non-ascii7 character is
 *   found (contributed by Andrew Worsley <epaanwo@asac.ericsson.se>).
 * - 0.11b (05/03/01) : Replaced control of all debug info by
 *   global flags.
 * - 0.12b (11/08/01) : In decode_pdu(), added a debug feature
 *   that would accept any numplan, print a warning and go ahead,
 *   e.g. for numplan #85 [reserved] (requested by Thomas Gutzler
 *   <giggel@hadiko.de>). Changed numplan processing to a
 *   switch structure to make it easier supporting more in the
 *   future.
 * - 0.13b (18/10/01) : Added formal support for numplans #83,
 *   #85, #89, #A1 and #D0. Kudos to Thomas Gutzler
 *   (<giggel@hadiko.de>) for his tremendous help in tracking
 *   that down. Modified the numplan switch cases to hex values.
 *   Made the <OA> byteswap conditional to <numplan> value.
 *   Fixed a bug where the msg->fromgsm string was not always
 *   properly terminated.
 *========================================================*/

#include <unistd.h>
#include <stdio.h>                         /* for fprintf */
#include <stdlib.h>                  /* for errno & stuff */
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <math.h>                            /* for pow() */

#include "sms_serv.h"
#include "gsmdev.h"
#include "pdu.h"

/*========================================================*/
/* For debugging purposes only - comment out for normal compile */
/* #define INCL_DEBUG_CODE */

/*========================================================*/
/**********           GLOBAL VARIABLES             ********/
/*========================================================*/
extern int debug;                    /* debug level/flags */

/*========================================================*/
/**********               FUNCTIONS                ********/
/*========================================================*/
inline int dapad (char *string)
{
  int len;
  
  len = strlen (string);
  if ((strlen (string) % 2) == 1) {
    /* str_insert (string, (len - 1), 'F'); */
    strcat (string, "F");
  }
}                                             /* dapad () */
/*========================================================*/
inline int byteswap (char *string)
{
  int len, i;
  char c;
  
  len = strlen (string);
  for (i = 1; i < len; i += 2) {
    c = string[i];
    string[i] = string[i - 1];
    string[i - 1] = c;
  }
}                                          /* byteswap () */
/*========================================================*/
int parse_csts_date (char *instr, char *outstr)
{
  char yy[3];
  char *errptr;
  int y, s;
  
  /* extract the "year" part */
  strncpy (yy, instr, 2);
  yy[2] = '\0';
  
  /* convert it to decimal */
  y = strtol (yy, &errptr, 10);
  if (*errptr != '\0') {
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "Invalid input date string format : [%s]\n", instr);
    }
    return (-1);
  }
  
  /* compute the required century */
  if (y > 95) {
    s = 19;
  }
  else {
    s = 20;
  }
  
  /* build output string */
  sprintf (outstr, "%d%s", s, instr);
  
  if (strlen (outstr) != 8) {
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "Invalid output date string format : [%s]\n", outstr);
    }
    return (-1);
  }
  else {
    return (strlen (outstr));
  }
}                                   /* parse_csts_date () */
/*========================================================*/
int parse_csts_time (char *instr, char *outstr)
{
  /* convert from hhmmssuu to xxhyymzzs */
  char hh[3];
  char mm[3];
  char ss[3];
  
  /* split instr into its components */
  strncpy (hh, instr, 2);
  hh[2] = '\0';
  shiftleft (instr, 2);
  
  strncpy (mm, instr, 2);
  mm[2] = '\0';
  shiftleft (instr, 2);
  
  strncpy (ss, instr, 2);
  ss[2] = '\0';
  
  /* build output string */
  sprintf (outstr, "%sh%sm%ss", hh, mm, ss);
  
  return (strlen (outstr));
}                                   /* parse_csts_time () */
/*========================================================*/
int ascii7topacked8 (char *in, char *out)
{
  int i, inbitlen, outbitlen, outfree, chunklen;
  unsigned char cur_c, out_c, mask, trans;
  char scratch[3];

  /*--------------------------------------Initializations */
  i = 0;
  outbitlen = 0;
  out_c = '\0';

  /*---------------------------------Main processing loop */
  while (in[i]) {
    cur_c = in[i];
    inbitlen = 7;

    while (inbitlen) {
      outfree = (8 - outbitlen);
      chunklen = min (outfree, inbitlen);
      if (!chunklen) {
        /* out_c is full - save it */
        sprintf (scratch, "%0.2X", out_c);
        strcat (out, scratch);
        /* re-initialize out_c */
        out_c = '\0';
        outbitlen = 0;
        chunklen = inbitlen;
      }                                 /* if (!chunklen) */
      /* save chunklen bits to trans */
      mask = (pow (2, chunklen) - 1);
      trans = cur_c & mask;
      /* offset what's left of cur_c for further use */
      cur_c = cur_c >> chunklen;
      /* offset trans for merge with out_c */
      trans = trans << outbitlen;
      /* merge trans in out_c */
      out_c |= trans;
      /* update length counters */
      inbitlen -= chunklen;
      outbitlen += chunklen;
    }                                 /* while (inbitlen) */

    i++;
  }                                      /* while (in[i]) */
  
  /*------------------------------------------Conclusions */
  /* save the remaining last bits */
  sprintf (scratch, "%0.2X", out_c);
  strcat (out, scratch);

  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "Message in packed8 format : [%s]\n", out);
  }
  return (strlen (out) / 2);
}                                   /* ascii7topacked8 () */
/*========================================================*/
int packed8toascii7 (char *in, char *out)
{
  int inbitlen, outbitlen, outfree, chunklen;
  unsigned char cur_c, out_c, mask, trans;
  char scratch[3];
  char *ptr;
  char *endptr;

  /*--------------------------------------Initializations */
  ptr = in;
  outbitlen = 0;
  out_c = '\0';

  while (ptr < (in + strlen (in))) {
    /* chop the in-string 2 (hex) char at a time */
    strncpy (scratch, ptr, 2);
    scratch[2] = '\0';
    /* convert hex digit pair to binary value */
    cur_c = strtol (scratch, &endptr, 16);
    inbitlen = 8;

    while (inbitlen) {
      outfree = (7 - outbitlen);
      chunklen = min (outfree, inbitlen);
      if (!chunklen) {
        /* out_c is full - save it */
	if (out_c & 0x80) {
          if (debug & DEBUG_PRINTF) {
            fprintf (stderr, "packed8toascii7() non-ascii7 character found! 0x%x\n",
	            out_c);
          }
	  return (-1);     /* non-ascii7 character found! */
	}
        sprintf (scratch, "%c", out_c);
        strcat (out, scratch);
        /* re-initialize out_c */
        out_c = '\0';
        outbitlen = 0;
        chunklen = min (inbitlen, 7);
      }                                 /* if (!chunklen) */
      /* save chunklen bits to trans */
      mask = (pow (2, chunklen) - 1);
      trans = cur_c & mask;
      /* offset what's left of cur_c for further use */
      cur_c = cur_c >> chunklen;
      /* offset trans for merge with out_c */
      trans = trans << outbitlen;
      /* merge trans in out_c */
      out_c |= trans;
      /* update length counters */
      inbitlen -= chunklen;
      outbitlen += chunklen;
    }                                 /* while (inbitlen) */

    ptr += 2;
  }                                  /* while (ptr < ...) */
  /* save the last bits */
  if (out_c & 0x80) {
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "packed8toascii7() trailing non-ascii7 character found! 0x%x\n",
	      out_c);
    }
    return (-1);           /* non-ascii7 character found! */
  }
  sprintf (scratch, "%c", out_c);
  strcat (out, scratch);

  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "Raw message after packed8toascii7() : [%s]\n", out);
  }
  return (strlen (out));
}                                   /* packed8toascii7 () */
/*========================================================*/

/*========================================================*/
/*#################### main PDU functions ################*/
/*========================================================*/
char *encode_pdu (struct symbols *symbols)
{
  /* This function is largely based on proof-of-concept code *
   * provided by Franois Baligant <francois@euronet.be>.    */
   
  char *pdu;
  char scratch[MAXPHNUMLEN + 1];
  unsigned int tpdu;
  int skip;
  char numplan[3];
  char dalen[3];
  char *buff;
  int tp_udl;
  
  /*---------------------------------------Initialization */
  pdu = (char *) malloc ((BUFFSIZE + 1) * sizeof (char));
  if (!pdu) {
    syslog ((FACILITY | LOG_ERR), "can't malloc() for PDU string.");
    return (NULL);
  }
  pdu[0] = '\0';
  
  /*--------------------------------------Field #1 : TPDU */
  /* tpdu = (TP_MTI_SMS_SUBMIT | TP_MMS_NO_MORE_MSG | TP_VPF_RELATIVE | TP_SRI_SR_NOT_REQD); */
  tpdu = (TP_MTI_SMS_SUBMIT | TP_MMS_NO_MORE_MSG);
  sprintf (scratch, "%0.2X", tpdu);
  strcat (pdu, scratch);
  
  /*----------------------------------------Field #2 : MR */
  /* Message Reference left NULL - should be set by the phone */
  strcat (pdu, "00");

  /*----------------------------------------Field #3 : DA */
  /* Destination Address : Address Length, Numbering Plan, Number */
  if (symbols->destgsm[0] == '+') {
    /* international number */
    skip = 1;
    strcpy (numplan, "91");
  }
  else if ((symbols->destgsm[0] == '0') && (symbols->destgsm[1] == '0')) {
    /* international number */
    skip = 2;
    strcpy (numplan, "91");
  }
  else {
    /* local number */
    skip = 0;
    strcpy (numplan, "81");
  }
  
  /* remove the international call prefix - if any */
  strcpy (scratch, symbols->destgsm);
  if (skip > 0) {
    shiftleft (scratch, skip);
  }
  
  /* Now store address length as hex value */
  sprintf (dalen, "%0.2X", strlen (scratch));
  strcat (pdu, dalen);
  
  /* Add the numbering plan */
  strcat (pdu, numplan);
  
  /* Finally append the padded, byte-swapped, destination */
  dapad (scratch);
  byteswap (scratch);
  strcat (pdu, scratch);

  /*---------------------------------------Field #4 : PID */
  sprintf (scratch, "%0.2X", PID_SMS);
  strcat (pdu, scratch);
  
  /*---------------------------------------Field #5 : DCS */
  sprintf (scratch, "%0.2X", DCS_PCCP437);
  strcat (pdu, scratch);
  
  /*----------------------------------------Field #6 : VP */
  if (tpdu & TP_VPF_RELATIVE) {
    strcat (pdu, "AA");                         /* 4 days */
  }
  
  /*----------------------------Field #7 : TP_UDL + TP_UD */
  /* TP_UDL is the length of TP_UD measured in chars (NOT in  */
  /* bytes). Thanks to  <eric.guidi@wavecom.fr>).             */
  buff = (char *) malloc (((2 * MAXMSGLEN) + 1) * sizeof (char));
  buff[0] = '\0';
  
  tp_udl = strlen (symbols->message);
  ascii7topacked8 (symbols->message, buff);

  sprintf (scratch, "%0.2X", tp_udl);
  strcat (pdu, scratch);
  strcat (pdu, buff);
  /*-------------------------------------------Conclusion */
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "PDU-encoded string : [%s]\n", pdu);
  }
  free (buff);
  return (pdu);
}                                        /* encode_pdu () */
/*========================================================*/
int decode_pdu (char *pdu, struct mbox_item *msg, cap_matrix capmatrix)
{
  char *scratch;
  char *ptr;
  char *errptr;
  int scalen, scalen_p, scalen_t;
  char numplan[3];
  int n_numplan;
  char sca[MAXPHNUMLEN + 1];
  int tpdu;
  int oalen, oalen_p;
  int pid;
  int dcs;
  char *datestr;
  char *timestr;
  int tpudl;

  /*---------------------------------------Initialization */
  scratch = (char *) malloc ((BUFFSIZE + 1) * sizeof (char));
  if (!scratch) {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't malloc() (scratch).");
    return (-1);
  }
  scratch[0] = '\0';
  
  datestr = (char *) malloc ((MINIBUFF + 1) * sizeof (char));
  if (!datestr) {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't malloc() (datestr).");
    return (-1);
  }
  datestr[0] = '\0';
  
  timestr = (char *) malloc ((MINIBUFF + 1) * sizeof (char));
  if (!timestr) {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't malloc() (timestr).");
    return (-1);
  }
  timestr[0] = '\0';
  ptr = pdu;
  n_numplan = 0;
  
  /*--------------------------------Field #0 : SCA (opt.) */
  /* only appears on specific GSM models */
  if (capmatrix.capflags & HAS_SCA_IN_PDU) {
    /* first extract <scalen> - 1 byte */
    /* Warning: <scalen> is measured in PDU "bytes", <numplan> and */
    /* padding char. included */
    if (strlen (ptr) >= 2) {
      strncpy (scratch, ptr, 2);
      scratch[2] = '\0';
    }
    else {
      syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse PDU (scalen).");
      return (-1);
    }
    /* convert it to decimal */
    scalen = strtol (scratch, &errptr, 16);
    if (*errptr != '\0') {
      syslog ((FACILITY | LOG_ERR), "decode_pdu() - invalid scalen value.");
      return (-1);
    }
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "SCALEN value : [%d]\n", scalen);
    }
    /* jump to next */
    ptr += 2;

    /* then extract the numbering plan - 1 byte */
    if (strlen (ptr) >= 2) {
      strncpy (numplan, ptr, 2);
      numplan[2] = '\0';
    }
    else {
      syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse PDU (sca numplan).");
      return (-1);
    }
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "SCA NUMPLAN value : [%s]\n", numplan);
    }
    /* jump to next */
    ptr += 2;

    /* now get the SCA number - (scalen - 1) * 2 char */
    scalen_p = ((scalen - 1) * 2);
    if (strlen (ptr) >= scalen_p) {
      strncpy (scratch, ptr, scalen_p);
      scratch[scalen_p] = '\0';
    }
    else {
      syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse PDU (sca).");
      return (-1);
    }
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "Raw SCA value : [%s]\n", scratch);
    }
    /* byteswap the SCA field */
    byteswap (scratch);
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "Byteswaped SCA value : [%s]\n", scratch);
    }
    /* correct scalen_p according to presence of padding char. */
    if (scratch[scalen_p - 1] == 'F') {
      /* pad char. present */
      scalen_t = scalen_p - 1;
    }
    else {
      /* no pad char. */
      scalen_t = scalen_p;
    }
    /* build the SCA ph. number */
    if (strcmp (numplan, "81") == 0) {
      /* local number - nothing special to do */
      strncpy (sca, scratch, scalen_t);
    }
    else if (strcmp (numplan, "91") == 0) {
      /* international number */
      strcpy (sca, "+");
      strncat (sca, scratch, scalen_t);
    }
    else {
      syslog ((FACILITY | LOG_ERR), "decode_pdu() - unsupported SCA numplan (%s).",
             numplan);
      return (-1);
    }
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "Final SCA value : [%s]\n", sca);
    }
    /* jump to next */
    ptr += scalen_p;
  }           /* if (capmatrix.capflags & HAS_SCA_IN_PDU) */
  
  /*--------------------------------------Field #1 : TPDU */
  /* extract the first byte */
  if (strlen (ptr) >= 2) {
    strncpy (scratch, ptr, 2);
    scratch[2] = '\0';
  }
  else {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse PDU (tpdu).");
    return (-1);
  }
  /* convert it to decimal */
  tpdu = strtol (scratch, &errptr, 16);
  if (*errptr != '\0') {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - invalid tpdu value.");
    return (-1);
  }
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "TPDU value : [%d]\n", tpdu);
  }
  /* analyse it - later */
  /* jump to next */
  ptr += 2;
  
  /*----------------------------------------Field #2 : OA */
  /*.......................first extract <oalen> - 1 byte */
  /* Warning: <oalen> is measured in characters, <numplan> and padding */
  /* character excluded */
  if (strlen (ptr) >= 2) {
    strncpy (scratch, ptr, 2);
    scratch[2] = '\0';
  }
  else {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse PDU (oalen).");
    return (-1);
  }
  /* convert it to decimal */
  oalen = strtol (scratch, &errptr, 16);
  if (*errptr != '\0') {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - invalid oalen value.");
    return (-1);
  }
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "OALEN value : [%d]\n", oalen);
  }
  /* jump to next */
  ptr += 2;
  
  /*.............then extract the numbering plan - 1 byte */
  if (strlen (ptr) >= 2) {
    strncpy (numplan, ptr, 2);
    numplan[2] = '\0';
  }
  else {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse PDU (numplan).");
    return (-1);
  }
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "NUMPLAN value : [%s]\n", numplan);
  }
  /* convert it for further use (switch) */
  n_numplan = strtol (numplan, (char **)NULL, 16);
  /* jump to next */
  ptr += 2;
  
  /*......now get the origin. number - oalen (+ 1 ?) char */
  if ((oalen % 2) == 1) {
    /* odd length => pad char. present */
    oalen_p = oalen + 1;
  }
  else {
    /* no pad char. */
    oalen_p = oalen;
  }
  if (strlen (ptr) >= oalen_p) {
    strncpy (scratch, ptr, oalen_p);
    scratch[oalen_p] = '\0';
  }
  else {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse PDU (oa).");
    return (-1);
  }
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "Raw OA value : [%s]\n", scratch);
  }

  /* byteswap the OA field when needed */
  if (n_numplan != 0xD0) {
    byteswap (scratch);
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "Byteswaped OA value : [%s]\n", scratch);
    }
  }
  else {
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "No byteswap needed for numplan [%s]\n", numplan);
    }
  }

  /* build and save the source ph. number according to <numplan> */
  switch (n_numplan) {
    case (0x81): {
      /* local number - nothing special to do */
      strncpy (msg->fromgsm, scratch, oalen);
      msg->fromgsm[oalen] = '\0';
      break;
    }

    case (0x91): {
      /* international number */
      strcpy (msg->fromgsm, "+");
      strncat (msg->fromgsm, scratch, oalen);
      msg->fromgsm[oalen + 1] = '\0';
      break;
    }

    case (0x83): {
      /* Data numbering plan (X.121) */
      if ((oalen > 2) && (scratch[0] == '0') && (scratch[0] == '0')) {
        /* International number w/ leading "00" - replace w/ '+' */
	strcpy (msg->fromgsm, "+");
	strncat (msg->fromgsm, (scratch + 2), (oalen - 2));
	msg->fromgsm[oalen - 1] = '\0';
      }
      else if ((oalen > 1) && (scratch[0] == '0')) {
        /* Single leading '0' -> local number - keep as is */
        strncpy (msg->fromgsm, scratch, oalen);
        msg->fromgsm[oalen] = '\0';
      }
      else {
        /* International number (?) */
	strcpy (msg->fromgsm, "+");
	strncat (msg->fromgsm, scratch, oalen);
	msg->fromgsm[oalen + 1] = '\0';
      }
      break;
    }

    case (0xA1): {
      /* National number - nothing special to do */
      strncpy (msg->fromgsm, scratch, oalen);
      msg->fromgsm[oalen] = '\0';
      break;
    }

    case (0xD0): {
      /* Alphanum. value in GSM TS 03.38 7bit alphabet */
      /* likely to contain the originating website name */
      packed8toascii7 (scratch, msg->fromgsm);
      break;
    }

    case (0x85):
    case (0x89): {
      /* reserved value or "private" numplan */
      /* <OA> field likely just garbage - nuke it */
      strcpy (msg->fromgsm, "000000000000");
      syslog ((FACILITY | LOG_NOTICE), "decode_pdu() - reserved numplan (%s), <OA> was [%s] - ignored.",
             numplan, scratch);
      break;
    }

    default: {
      if (debug & DEBUG_PDU) {
	/* unusual value, assumed local - nothing special to do */
	strncpy (msg->fromgsm, scratch, oalen);
        msg->fromgsm[oalen] = '\0';
	/* log a warning */
	syslog ((FACILITY | LOG_NOTICE), "decode_pdu() - unknown numplan (%s), assumed local.",
               numplan);
	/* print a debug message */
	fprintf (stderr, "Unknown numplan (%s), assumed local. PLEASE INVESTIGATE !\n",
	        numplan);
	fprintf (stderr, "Originating phone number: <%s>\n", msg->fromgsm);
	fprintf (stderr, "Proceeding anyway.\n");
	break;
      }
      else {
	syslog ((FACILITY | LOG_ERR), "decode_pdu() - unsupported numplan (%s).",
               numplan);
	return (-1);
      }
    }
  }                                 /* switch (n_numplan) */
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "Final OA value : [%s]\n", msg->fromgsm);
  }
  /* jump to next */
  ptr += oalen_p;

  /*---------------------------------------Field #3 : PID */
  /* extract the <pid> field - 1 byte */
  if (strlen (ptr) >= 2) {
    strncpy (scratch, ptr, 2);
    scratch[2] = '\0';
  }
  else {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse PDU (pid).");
    return (-1);
  }
  /* convert it to decimal */
  pid = strtol (scratch, &errptr, 16);
  if (*errptr != '\0') {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - invalid pid value (%s).",
            scratch);
    return (-1);
  }
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "PID value : [%d]\n", pid);
  }
  /* analyse it - later */
  /* jump to next */
  ptr += 2;

  /*---------------------------------------Field #4 : DCS */
  /* extract the <dcs> field - 1 byte */
  if (strlen (ptr) >= 2) {
    strncpy (scratch, ptr, 2);
    scratch[2] = '\0';
  }
  else {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse PDU (dcs).");
    return (-1);
  }
  /* convert it to decimal */
  dcs = strtol (scratch, &errptr, 16);
  if (*errptr != '\0') {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - invalid dcs value (%s).",
            scratch);
    return (-1);
  }
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "DCS value : [%d]\n", dcs);
  }
  /* analyse it - later */
  /* jump to next */
  ptr += 2;

  /*--------------------------------------Field #5 : CSTS */
  /* extract the <csts> field - 7 byte */
  if (strlen (ptr) >= 14) {
    strncpy (scratch, ptr, 14);
    scratch[14] = '\0';
  }
  else {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse PDU (csts).");
    return (-1);
  }
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "Raw CSTS value : [%s]\n", scratch);
  }
  /* byteswap the CSTS field */
  byteswap (scratch);
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "Byteswaped CSTS value : [%s]\n", scratch);
  }
  /* now split the byteswaped <csts> string into raw date and time */
  strncpy (datestr, scratch, 6);
  datestr[6] = '\0';
  shiftleft (scratch, 6);
  strcpy (timestr, scratch);
  /* parse each of them */
  if (parse_csts_date (datestr, msg->date) == -1) {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse <csts> date.");
    return (-1);
  }
  if (parse_csts_time (timestr, msg->time) == -1) {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse <csts> time.");
    return (-1);
  }
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "message date = [%s]\n", msg->date);
    fprintf (stderr, "message time = [%s]\n", msg->time);
  }

  /* jump to next */
  ptr += 14;

  /*-------------------------------------Field #6 : TP-UD */
  /* first extract <tp-udl> - 1 byte */
  if (strlen (ptr) >= 2) {
    strncpy (scratch, ptr, 2);
    scratch[2] = '\0';
  }
  else {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - can't parse PDU (tp-udl).");
    return (-1);
  }
  /* convert it to decimal */
  tpudl = strtol (scratch, &errptr, 16);
  if (*errptr != '\0') {
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - invalid tp-udl value.");
    return (-1);
  }
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "TP-UDL value : [%d]\n", tpudl);
  }
  /* jump to next */
  ptr += 2;
  
  /* the rest is TP-UD - decode it */
  strcpy (scratch, ptr);
  if (packed8toascii7 (scratch, msg->text) != tpudl) {
    if (debug & DEBUG_PRINTF) {
      fprintf (stderr, "Decoded TP-UD : [%s] length: %d (corruption ?)\n",
              msg->text, strlen (msg->text));
    }
    syslog ((FACILITY | LOG_ERR), "decode_pdu() - corrupt data, <tpudl> don't match.");
    return (-1);
  }
  if (debug & DEBUG_PRINTF) {
    fprintf (stderr, "Decoded TP-UD string : [%s]\n", msg->text);
  }

  /*-------------------------------------------Conclusion */
  free (scratch);
  free (datestr);
  free (timestr);
  return (strlen (msg->text));
}                                        /* decode_pdu () */
/*========================================================*/

/*==========================================================
 * EOF : pdu.c
 *===================*/
